@@ -284,7 +284,7 @@ struct _GpgCtx {
 	unsigned int complete:1;
 	unsigned int seen_eof1:1;
 	unsigned int seen_eof2:1;
-	unsigned int flushed:1;      /* flushed the diagnostics stream (ie, what we read from stderr) */
+	unsigned int flushed:1;      /* flushed the diagnostics stream (aka stderr) */
 	unsigned int always_trust:1;
 	unsigned int armor:1;
 	unsigned int need_passwd:1;
@@ -313,7 +313,7 @@ gpg_ctx_new (GMimeSession *session, const char *path)
 	gpg->mode = GPG_CTX_MODE_SIGN;
 	gpg->session = session;
 	g_object_ref (session);
-	gpg->userid_hint = g_hash_table_new (g_str_hash, g_str_equal);
+	gpg->userid_hint = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
 	gpg->complete = FALSE;
 	gpg->seen_eof1 = TRUE;
 	gpg->seen_eof2 = FALSE;
@@ -469,13 +469,6 @@ gpg_ctx_get_diagnostics (struct _GpgCtx *gpg)
 	return (const char *) gpg->diag->data;
 }
 
-static void
-userid_hint_free (gpointer key, gpointer value, gpointer user_data)
-{
-	g_free (key);
-	g_free (value);
-}
-
 static void
 gpg_ctx_free (struct _GpgCtx *gpg)
 {
@@ -485,7 +478,6 @@ gpg_ctx_free (struct _GpgCtx *gpg)
 	if (gpg->session)
 		g_object_unref (gpg->session);
 	
-	g_hash_table_foreach (gpg->userid_hint, userid_hint_free, NULL);
 	g_hash_table_destroy (gpg->userid_hint);
 	
 	g_free (gpg->path);
@@ -549,8 +541,18 @@ gpg_hash_str (GMimeCipherHash hash)
 		return "--digest-algo=MD5";
 	case GMIME_CIPHER_HASH_SHA1:
 		return "--digest-algo=SHA1";
+	case GMIME_CIPHER_HASH_SHA224:
+		return "--digest-algo=SHA224";
+	case GMIME_CIPHER_HASH_SHA256:
+		return "--digest-algo=SHA256";
+	case GMIME_CIPHER_HASH_SHA384:
+		return "--digest-algo=SHA384";
+	case GMIME_CIPHER_HASH_SHA512:
+		return "--digest-algo=SHA512";
 	case GMIME_CIPHER_HASH_RIPEMD160:
 		return "--digest-algo=RIPEMD160";
+	case GMIME_CIPHER_HASH_TIGER192:
+		return "--digest-algo=TIGER192";
 	default:
 		return NULL;
 	}
@@ -1102,8 +1104,30 @@ gpg_ctx_parse_status (struct _GpgCtx *gpg, GError **err)
 	} else {
 		switch (gpg->mode) {
 		case GPG_CTX_MODE_SIGN:
-			if (!strncmp (status, "SIG_CREATED ", 12)) {
-				/* FIXME: save this state? */
+			if (strncmp (status, "SIG_CREATED ", 12) != 0)
+				break;
+			
+			status += 12;
+			
+			/* skip the next single-char token ("D" for detached) */
+			status = next_token (status, NULL);
+			
+			/* skip the public-key algo token */
+			status = next_token (status, NULL);
+			
+			/* this token is the hash algorithm used */
+			switch (strtol (status, NULL, 10)) {
+			case 1: gpg->hash = GMIME_CIPHER_HASH_MD5; break;
+			case 2: gpg->hash = GMIME_CIPHER_HASH_SHA1; break;
+			case 3:	gpg->hash = GMIME_CIPHER_HASH_RIPEMD160; break;
+			case 5: gpg->hash = GMIME_CIPHER_HASH_MD2; break; /* ? */
+			case 6: gpg->hash = GMIME_CIPHER_HASH_TIGER192; break; /* ? */
+			case 7: gpg->hash = GMIME_CIPHER_HASH_HAVAL5160; break; /* ? */
+			case 8: gpg->hash = GMIME_CIPHER_HASH_SHA256; break;
+			case 9: gpg->hash = GMIME_CIPHER_HASH_SHA384; break;
+			case 10: gpg->hash = GMIME_CIPHER_HASH_SHA512; break;
+			case 11: gpg->hash = GMIME_CIPHER_HASH_SHA224; break;
+			default: break;
 			}
 			break;
 		case GPG_CTX_MODE_VERIFY:
@@ -1156,19 +1180,31 @@ gpg_ctx_parse_status (struct _GpgCtx *gpg, GError **err)
 	return 0;
 }
 
+static size_t
+nearest_pow (size_t num)
+{
+	size_t n = num > 0 ? num - 1 : 0;
+	
+	n |= n >> 1;
+	n |= n >> 2;
+	n |= n >> 4;
+	n |= n >> 8;
+	n |= n >> 16;
+	n++;
+	
+	return n;
+}
+
 #define status_backup(gpg, start, len) G_STMT_START {                     \
 	if (gpg->statusleft <= len) {                                     \
-		unsigned int slen, soff;                                  \
+		size_t slen, soff;                                        \
 		                                                          \
-		slen = soff = gpg->statusptr - gpg->statusbuf;            \
-		slen = slen ? slen : 1;                                   \
+		soff = gpg->statusptr - gpg->statusbuf;                   \
+		slen = nearest_pow (soff + len + 1);                      \
 		                                                          \
-		while (slen < soff + len)                                 \
-			slen <<= 1;                                       \
-		                                                          \
-		gpg->statusbuf = g_realloc (gpg->statusbuf, slen + 1);    \
+		gpg->statusbuf = g_realloc (gpg->statusbuf, slen);        \
 		gpg->statusptr = gpg->statusbuf + soff;                   \
-		gpg->statusleft = slen - soff;                            \
+		gpg->statusleft = (slen - 1) - soff;                      \
 	}                                                                 \
 	                                                                  \
 	memcpy (gpg->statusptr, start, len);                              \
@@ -1539,9 +1575,12 @@ gpg_sign (GMimeCipherContext *context, const char *userid, GMimeCipherHash hash,
 		return -1;
 	}
 	
+	/* save the hash used */
+	hash = gpg->hash;
+	
 	gpg_ctx_free (gpg);
 	
-	return 0;
+	return hash;
 }
 
 
