@@ -58,19 +58,26 @@ static gboolean message_remove_header (GMimeObject *object, const char *header);
 static char *message_get_headers (GMimeObject *object);
 static ssize_t message_write_to_stream (GMimeObject *object, GMimeStream *stream);
 
-static void message_add_recipients_from_string (GMimeMessage *message, GMimeRecipientType type, const char *str);
-
 static ssize_t write_structured (GMimeStream *stream, const char *name, const char *value);
 static ssize_t write_addrspec (GMimeStream *stream, const char *name, const char *value);
 static ssize_t write_received (GMimeStream *stream, const char *name, const char *value);
 static ssize_t write_subject (GMimeStream *stream, const char *name, const char *value);
 static ssize_t write_msgid (GMimeStream *stream, const char *name, const char *value);
 
+static void to_list_changed (InternetAddressList *list, GMimeMessage *message);
+static void cc_list_changed (InternetAddressList *list, GMimeMessage *message);
+static void bcc_list_changed (InternetAddressList *list, GMimeMessage *message);
+
 
 static GMimeObjectClass *parent_class = NULL;
 
-static const char *recipient_types[] = {
-	"To", "Cc", "Bcc"
+static struct {
+	const char *name;
+	GCallback changed_cb;
+} recipient_types[] = {
+	{ "To",  G_CALLBACK (to_list_changed)  },
+	{ "Cc",  G_CALLBACK (cc_list_changed)  },
+	{ "Bcc", G_CALLBACK (bcc_list_changed) }
 };
 
 #define N_RECIPIENT_TYPES G_N_ELEMENTS (recipient_types)
@@ -132,20 +139,88 @@ g_mime_message_class_init (GMimeMessageClass *klass)
 	object_class->write_to_stream = message_write_to_stream;
 }
 
+static gulong
+lookup_recipient_changed_id (GMimeMessage *message, GMimeRecipientType type)
+{
+	InternetAddressList *list;
+	GCallback changed_cb;
+	const char *name;
+	
+	changed_cb = recipient_types[type].changed_cb;
+	name = recipient_types[type].name;
+	
+	list = g_hash_table_lookup (message->recipients, name);
+	
+	return g_signal_handler_find (list, G_SIGNAL_MATCH_FUNC | G_SIGNAL_MATCH_DATA,
+				      0, 0, NULL, changed_cb, message);
+}
+
+static void
+connect_changed_event (GMimeMessage *message, GMimeRecipientType type)
+{
+	InternetAddressList *list;
+	
+	list = g_hash_table_lookup (message->recipients, recipient_types[type].name);
+	
+	g_signal_connect (list, "changed", recipient_types[type].changed_cb, message);
+}
+
+static void
+disconnect_changed_event (GMimeMessage *message, GMimeRecipientType type)
+{
+	InternetAddressList *list;
+	gulong id;
+	
+	list = g_hash_table_lookup (message->recipients, recipient_types[type].name);
+	id = lookup_recipient_changed_id (message, type);
+	g_signal_handler_disconnect (list, id);
+}
+
+static void
+block_changed_event (GMimeMessage *message, GMimeRecipientType type)
+{
+	InternetAddressList *list;
+	gulong id;
+	
+	list = g_hash_table_lookup (message->recipients, recipient_types[type].name);
+	id = lookup_recipient_changed_id (message, type);
+	g_signal_handler_block (list, id);
+}
+
+static void
+unblock_changed_event (GMimeMessage *message, GMimeRecipientType type)
+{
+	InternetAddressList *list;
+	gulong id;
+	
+	list = g_hash_table_lookup (message->recipients, recipient_types[type].name);
+	id = lookup_recipient_changed_id (message, type);
+	g_signal_handler_unblock (list, id);
+}
+
 static void
 g_mime_message_init (GMimeMessage *message, GMimeMessageClass *klass)
 {
 	GMimeHeaderList *headers = ((GMimeObject *) message)->headers;
+	InternetAddressList *list;
+	guint i;
 	
 	message->from = NULL;
 	message->reply_to = NULL;
-	message->recipients = g_hash_table_new (g_str_hash, g_str_equal);
+	message->recipients = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, g_object_unref);
 	message->subject = NULL;
 	message->date = 0;
 	message->tz_offset = 0;
 	message->message_id = NULL;
 	message->mime_part = NULL;
 	
+	/* initialize recipient lists */
+	for (i = 0; i < N_RECIPIENT_TYPES; i++) {
+		list = internet_address_list_new ();
+		g_hash_table_insert (message->recipients, (char *) recipient_types[i].name, list);
+		connect_changed_event (message, i);
+	}
+	
 	g_mime_header_list_register_writer (headers, "Sender", write_addrspec);
 	g_mime_header_list_register_writer (headers, "From", write_addrspec);
 	g_mime_header_list_register_writer (headers, "To", write_addrspec);
@@ -164,27 +239,19 @@ g_mime_message_init (GMimeMessage *message, GMimeMessageClass *klass)
 	g_mime_header_list_register_writer (headers, "References", write_structured);
 }
 
-static gboolean
-recipients_destroy (gpointer key, gpointer value, gpointer user_data)
-{
-	InternetAddressList *recipients = value;
-	
-	if (recipients)
-		internet_address_list_destroy (recipients);
-	
-	return TRUE;
-}
-
 static void
 g_mime_message_finalize (GObject *object)
 {
 	GMimeMessage *message = (GMimeMessage *) object;
+	guint i;
 	
 	g_free (message->from);
 	g_free (message->reply_to);
 	
-	/* destroy all recipients */
-	g_hash_table_foreach_remove (message->recipients, recipients_destroy, NULL);
+	/* disconnect changed handlers */
+	for (i = 0; i < N_RECIPIENT_TYPES; i++)
+		disconnect_changed_event (message, i);
+	
 	g_hash_table_destroy (message->recipients);
 	
 	g_free (message->subject);
@@ -584,7 +651,7 @@ write_addrspec (GMimeStream *stream, const char *name, const char *value)
 	
 	if (value && (addrlist = internet_address_list_parse_string (value))) {
 		internet_address_list_writer (addrlist, str);
-		internet_address_list_destroy (addrlist);
+		g_object_unref (addrlist);
 	}
 	
 	g_string_append_c (str, '\n');
@@ -595,6 +662,18 @@ write_addrspec (GMimeStream *stream, const char *name, const char *value)
 	return n;
 }
 
+static void
+message_add_recipients_from_string (GMimeMessage *message, GMimeRecipientType type, const char *str)
+{
+	InternetAddressList *recipients, *addrlist;
+	
+	recipients = g_mime_message_get_recipients (message, type);
+	
+	if ((addrlist = internet_address_list_parse_string (str))) {
+		internet_address_list_concat (recipients, addrlist);
+		g_object_unref (addrlist);
+	}
+}
 
 enum {
 	HEADER_FROM,
@@ -626,6 +705,7 @@ process_header (GMimeObject *object, const char *header, const char *value)
 	InternetAddressList *addrlist;
 	int offset, i;
 	time_t date;
+	gulong id;
 	
 	for (i = 0; i < G_N_ELEMENTS (message_headers); i++) {
 		if (!g_ascii_strcasecmp (message_headers[i], header))
@@ -637,7 +717,7 @@ process_header (GMimeObject *object, const char *header, const char *value)
 		g_free (message->from);
 		if ((addrlist = internet_address_list_parse_string (value))) {
 			message->from = internet_address_list_to_string (addrlist, FALSE);
-			internet_address_list_destroy (addrlist);
+			g_object_unref (addrlist);
 		} else {
 			message->from = NULL;
 		}
@@ -646,19 +726,25 @@ process_header (GMimeObject *object, const char *header, const char *value)
 		g_free (message->reply_to);
 		if ((addrlist = internet_address_list_parse_string (value))) {
 			message->reply_to = internet_address_list_to_string (addrlist, FALSE);
-			internet_address_list_destroy (addrlist);
+			g_object_unref (addrlist);
 		} else {
 			message->reply_to = NULL;
 		}
 		break;
 	case HEADER_TO:
+		block_changed_event (message, GMIME_RECIPIENT_TYPE_TO);
 		message_add_recipients_from_string (message, GMIME_RECIPIENT_TYPE_TO, value);
+		unblock_changed_event (message, GMIME_RECIPIENT_TYPE_TO);
 		break;
 	case HEADER_CC:
+		block_changed_event (message, GMIME_RECIPIENT_TYPE_CC);
 		message_add_recipients_from_string (message, GMIME_RECIPIENT_TYPE_CC, value);
+		unblock_changed_event (message, GMIME_RECIPIENT_TYPE_CC);
 		break;
 	case HEADER_BCC:
+		block_changed_event (message, GMIME_RECIPIENT_TYPE_BCC);
 		message_add_recipients_from_string (message, GMIME_RECIPIENT_TYPE_BCC, value);
+		unblock_changed_event (message, GMIME_RECIPIENT_TYPE_BCC);
 		break;
 	case HEADER_SUBJECT:
 		g_free (message->subject);
@@ -801,22 +887,25 @@ message_remove_header (GMimeObject *object, const char *header)
 		message->reply_to = NULL;
 		break;
 	case HEADER_TO:
-		type = recipient_types[GMIME_RECIPIENT_TYPE_TO];
+		type = recipient_types[GMIME_RECIPIENT_TYPE_TO].name;
+		block_changed_event (message, GMIME_RECIPIENT_TYPE_TO);
 		addrlist = g_hash_table_lookup (message->recipients, type);
-		g_hash_table_remove (message->recipients, type);
-		internet_address_list_destroy (addrlist);
+		internet_address_list_clear (addrlist);
+		unblock_changed_event (message, GMIME_RECIPIENT_TYPE_TO);
 		break;
 	case HEADER_CC:
-		type = recipient_types[GMIME_RECIPIENT_TYPE_CC];
+		type = recipient_types[GMIME_RECIPIENT_TYPE_CC].name;
+		block_changed_event (message, GMIME_RECIPIENT_TYPE_CC);
 		addrlist = g_hash_table_lookup (message->recipients, type);
-		g_hash_table_remove (message->recipients, type);
-		internet_address_list_destroy (addrlist);
+		internet_address_list_clear (addrlist);
+		unblock_changed_event (message, GMIME_RECIPIENT_TYPE_CC);
 		break;
 	case HEADER_BCC:
-		type = recipient_types[GMIME_RECIPIENT_TYPE_BCC];
+		type = recipient_types[GMIME_RECIPIENT_TYPE_BCC].name;
+		block_changed_event (message, GMIME_RECIPIENT_TYPE_BCC);
 		addrlist = g_hash_table_lookup (message->recipients, type);
-		g_hash_table_remove (message->recipients, type);
-		internet_address_list_destroy (addrlist);
+		internet_address_list_clear (addrlist);
+		unblock_changed_event (message, GMIME_RECIPIENT_TYPE_BCC);
 		break;
 	case HEADER_SUBJECT:
 		g_free (message->subject);
@@ -859,7 +948,8 @@ message_get_headers (GMimeObject *object)
 	} else {
 		g_mime_header_list_write_to_stream (object->headers, stream);
 		if (message->mime_part) {
-			g_mime_stream_write_string (stream, "MIME-Version: 1.0\n");
+			if (g_mime_object_get_header (message->mime_part, "Content-Type"))
+				g_mime_stream_write_string (stream, "MIME-Version: 1.0\n");
 			g_mime_header_list_write_to_stream (message->mime_part->headers, stream);
 		}
 	}
@@ -961,13 +1051,16 @@ g_mime_message_set_sender (GMimeMessage *message, const char *sender)
 	
 	g_free (message->from);
 	
-	addrlist = internet_address_list_parse_string (sender);
-	message->from = internet_address_list_to_string (addrlist, FALSE);
-	encoded = internet_address_list_to_string (addrlist, TRUE);
-	internet_address_list_destroy (addrlist);
-	
-	g_mime_header_list_set (GMIME_OBJECT (message)->headers, "From", encoded);
-	g_free (encoded);
+	if ((addrlist = internet_address_list_parse_string (sender))) {
+		message->from = internet_address_list_to_string (addrlist, FALSE);
+		encoded = internet_address_list_to_string (addrlist, TRUE);
+		g_mime_header_list_set (GMIME_OBJECT (message)->headers, "From", encoded);
+		g_object_unref (addrlist);
+		g_free (encoded);
+	} else {
+		g_mime_header_list_set (GMIME_OBJECT (message)->headers, "From", "");
+		message->from = NULL;
+	}
 }
 
 
@@ -1028,17 +1121,38 @@ g_mime_message_get_reply_to (GMimeMessage *message)
 static void
 sync_recipient_header (GMimeMessage *message, GMimeRecipientType type)
 {
-	const char *name = recipient_types[type];
-	const InternetAddressList *recipients;
+	GMimeObject *object = (GMimeObject *) message;
+	const char *name = recipient_types[type].name;
+	InternetAddressList *list;
 	char *string;
 	
 	/* sync the specified recipient header */
-	if ((recipients = g_mime_message_get_recipients (message, type))) {
-		string = internet_address_list_to_string (recipients, TRUE);
-		g_mime_header_list_set (GMIME_OBJECT (message)->headers, name, string);
+	if ((list = g_mime_message_get_recipients (message, type))) {
+		string = internet_address_list_to_string (list, TRUE);
+		g_mime_header_list_set (object->headers, name, string);
 		g_free (string);
-	} else
-		g_mime_header_list_set (GMIME_OBJECT (message)->headers, name, NULL);
+	} else {
+		/* list should never be NULL... */
+		g_mime_header_list_set (object->headers, name, NULL);
+	}
+}
+
+static void
+to_list_changed (InternetAddressList *list, GMimeMessage *message)
+{
+	sync_recipient_header (message, GMIME_RECIPIENT_TYPE_TO);
+}
+
+static void
+cc_list_changed (InternetAddressList *list, GMimeMessage *message)
+{
+	sync_recipient_header (message, GMIME_RECIPIENT_TYPE_CC);
+}
+
+static void
+bcc_list_changed (InternetAddressList *list, GMimeMessage *message)
+{
+	sync_recipient_header (message, GMIME_RECIPIENT_TYPE_BCC);
 }
 
 
@@ -1046,7 +1160,7 @@ sync_recipient_header (GMimeMessage *message, GMimeRecipientType type)
  * g_mime_message_add_recipient:
  * @message: MIME Message to change
  * @type: A #GMimeRecipientType
- * @name: The recipient's name
+ * @name: The recipient's name (or %NULL)
  * @address: The recipient's address
  *
  * Add a recipient of a chosen type to the MIME Message.
@@ -1059,65 +1173,12 @@ g_mime_message_add_recipient (GMimeMessage *message, GMimeRecipientType type, co
 	
 	g_return_if_fail (GMIME_IS_MESSAGE (message));
 	g_return_if_fail (type < N_RECIPIENT_TYPES);
-	g_return_if_fail (name != NULL);
 	g_return_if_fail (address != NULL);
 	
-	ia = internet_address_new_name (name, address);
-	
-	if ((recipients = g_hash_table_lookup (message->recipients, recipient_types[type])))
-		g_hash_table_remove (message->recipients, recipient_types[type]);
-	else
-		recipients = internet_address_list_new ();
-	
+	recipients = g_hash_table_lookup (message->recipients, recipient_types[type].name);
+	ia = internet_address_mailbox_new (name, address);
 	internet_address_list_add (recipients, ia);
-	internet_address_unref (ia);
-	
-	g_hash_table_insert (message->recipients, (char *) recipient_types[type], recipients);
-	sync_recipient_header (message, type);
-}
-
-
-static void
-message_add_recipients_from_string (GMimeMessage *message, GMimeRecipientType type, const char *str)
-{
-	InternetAddressList *recipients, *addrlist;
-	
-	if ((recipients = g_hash_table_lookup (message->recipients, recipient_types[type])))
-		g_hash_table_remove (message->recipients, recipient_types[type]);
-	
-	if ((addrlist = internet_address_list_parse_string (str))) {
-		if (recipients != NULL) {
-			internet_address_list_concat (recipients, addrlist);
-			internet_address_list_destroy (addrlist);
-		} else {
-			recipients = addrlist;
-		}
-	}
-	
-	if (recipients)
-		g_hash_table_insert (message->recipients, (char *) recipient_types[type], recipients);
-}
-
-
-/**
- * g_mime_message_add_recipients_from_string:
- * @message: MIME Message
- * @type: A #GMimeRecipientType
- * @str: A string of recipient names and addresses.
- *
- * Add a list of recipients of a chosen type to the MIME Message. The
- * string @str must be in the format specified in rfc822.
- **/
-void
-g_mime_message_add_recipients_from_string (GMimeMessage *message, GMimeRecipientType type, const char *str)
-{
-	g_return_if_fail (GMIME_IS_MESSAGE (message));
-	g_return_if_fail (type < N_RECIPIENT_TYPES);
-	g_return_if_fail (str != NULL);
-	
-	message_add_recipients_from_string (message, type, str);
-	
-	sync_recipient_header (message, type);
+	g_object_unref (ia);
 }
 
 
@@ -1131,13 +1192,17 @@ g_mime_message_add_recipients_from_string (GMimeMessage *message, GMimeRecipient
  * Returns: a list of recipients of a chosen type from the MIME
  * Message.
  **/
-const InternetAddressList *
+InternetAddressList *
 g_mime_message_get_recipients (GMimeMessage *message, GMimeRecipientType type)
 {
+	InternetAddressList *list;
+	
 	g_return_val_if_fail (GMIME_IS_MESSAGE (message), NULL);
 	g_return_val_if_fail (type < N_RECIPIENT_TYPES, NULL);
 	
-	return g_hash_table_lookup (message->recipients, recipient_types[type]);
+	list = g_hash_table_lookup (message->recipients, recipient_types[type].name);
+	
+	return list;
 }
 
 
@@ -1159,7 +1224,10 @@ g_mime_message_get_all_recipients (GMimeMessage *message)
 	g_return_val_if_fail (GMIME_IS_MESSAGE (message), NULL);
 	
 	for (i = 0; i < G_N_ELEMENTS (recipient_types); i++) {
-		if (!(recipients = g_hash_table_lookup (message->recipients, recipient_types[i])))
+		if (!(recipients = g_hash_table_lookup (message->recipients, recipient_types[i].name)))
+			continue;
+		
+		if (internet_address_list_length (recipients) == 0)
 			continue;
 		
 		if (list == NULL)
@@ -1360,8 +1428,6 @@ g_mime_message_get_mime_part (GMimeMessage *message)
 	if (message->mime_part == NULL)
 		return NULL;
 	
-	g_object_ref (message->mime_part);
-	
 	return message->mime_part;
 }
 
@@ -1392,21 +1458,21 @@ g_mime_message_set_mime_part (GMimeMessage *message, GMimeObject *mime_part)
 
 
 /**
- * g_mime_message_foreach_part:
- * @message: MIME message
+ * g_mime_message_foreach:
+ * @message: a #GMimeMessage
  * @callback: function to call on each of the mime parts contained by the mime message
- * @data: extra data to pass to the callback
+ * @user_data: user-supplied callback data
  *
- * Calls @callback on each of the mime parts in the mime message.
+ * Recursively calls @callback on each of the mime parts in the mime message.
  **/
 void
-g_mime_message_foreach_part (GMimeMessage *message, GMimePartFunc callback, gpointer data)
+g_mime_message_foreach (GMimeMessage *message, GMimeObjectForeachFunc callback, gpointer user_data)
 {
 	g_return_if_fail (GMIME_IS_MESSAGE (message));
 	g_return_if_fail (callback != NULL);
 	
+	callback ((GMimeObject *) message, message->mime_part, user_data);
+	
 	if (GMIME_IS_MULTIPART (message->mime_part))
-		g_mime_multipart_foreach (GMIME_MULTIPART (message->mime_part), callback, data);
-	else
-		callback (message->mime_part, data);
+		g_mime_multipart_foreach ((GMimeMultipart *) message->mime_part, callback, user_data);
 }
