@@ -119,7 +119,6 @@ g_mime_multipart_encrypted_class_init (GMimeMultipartEncryptedClass *klass)
 static void
 g_mime_multipart_encrypted_init (GMimeMultipartEncrypted *mpe, GMimeMultipartEncryptedClass *klass)
 {
-	mpe->protocol = NULL;
 	mpe->decrypted = NULL;
 	mpe->validity = NULL;
 }
@@ -129,8 +128,6 @@ g_mime_multipart_encrypted_finalize (GObject *object)
 {
 	GMimeMultipartEncrypted *mpe = (GMimeMultipartEncrypted *) object;
 	
-	g_free (mpe->protocol);
-	
 	if (mpe->decrypted)
 		g_object_unref (mpe->decrypted);
 	
@@ -173,13 +170,6 @@ multipart_encrypted_remove_header (GMimeObject *object, const char *header)
 static void
 multipart_encrypted_set_content_type (GMimeObject *object, GMimeContentType *content_type)
 {
-	GMimeMultipartEncrypted *mpe = (GMimeMultipartEncrypted *) object;
-	const char *protocol;
-	
-	protocol = g_mime_content_type_get_parameter (content_type, "protocol");
-	g_free (mpe->protocol);
-	mpe->protocol = g_strdup (protocol);
-	
 	GMIME_OBJECT_CLASS (parent_class)->set_content_type (object, content_type);
 }
 
@@ -207,12 +197,13 @@ GMimeMultipartEncrypted *
 g_mime_multipart_encrypted_new (void)
 {
 	GMimeMultipartEncrypted *multipart;
-	GMimeContentType *type;
+	GMimeContentType *content_type;
 	
 	multipart = g_object_new (GMIME_TYPE_MULTIPART_ENCRYPTED, NULL);
 	
-	type = g_mime_content_type_new ("multipart", "encrypted");
-	g_mime_object_set_content_type (GMIME_OBJECT (multipart), type);
+	content_type = g_mime_content_type_new ("multipart", "encrypted");
+	g_mime_object_set_content_type (GMIME_OBJECT (multipart), content_type);
+	g_object_unref (content_type);
 	
 	return multipart;
 }
@@ -281,7 +272,10 @@ g_mime_multipart_encrypted_encrypt (GMimeMultipartEncrypted *mpe, GMimeObject *c
 	g_mime_stream_reset (ciphertext);
 	
 	/* construct the version part */
-	version_part = g_mime_part_new ();
+	content_type = g_mime_content_type_new_from_string (ctx->encrypt_protocol);
+	version_part = g_mime_part_new_with_type (content_type->type, content_type->subtype);
+	g_object_unref (content_type);
+	
 	content_type = g_mime_content_type_new_from_string (ctx->encrypt_protocol);
 	g_mime_object_set_content_type (GMIME_OBJECT (version_part), content_type);
 	g_mime_part_set_content_encoding (version_part, GMIME_CONTENT_ENCODING_7BIT);
@@ -291,7 +285,6 @@ g_mime_multipart_encrypted_encrypt (GMimeMultipartEncrypted *mpe, GMimeObject *c
 	g_object_unref (wrapper);
 	g_object_unref (stream);
 	
-	mpe->protocol = g_strdup (ctx->encrypt_protocol);
 	mpe->decrypted = content;
 	g_object_ref (content);
 	
@@ -305,13 +298,14 @@ g_mime_multipart_encrypted_encrypt (GMimeMultipartEncrypted *mpe, GMimeObject *c
 	
 	/* save the version and encrypted parts */
 	/* FIXME: make sure there aren't any other parts?? */
-	g_mime_multipart_add_part (GMIME_MULTIPART (mpe), GMIME_OBJECT (version_part));
-	g_mime_multipart_add_part (GMIME_MULTIPART (mpe), GMIME_OBJECT (encrypted_part));
+	g_mime_multipart_add (GMIME_MULTIPART (mpe), GMIME_OBJECT (version_part));
+	g_mime_multipart_add (GMIME_MULTIPART (mpe), GMIME_OBJECT (encrypted_part));
 	g_object_unref (encrypted_part);
 	g_object_unref (version_part);
 	
 	/* set the content-type params for this multipart/encrypted part */
-	g_mime_object_set_content_type_parameter (GMIME_OBJECT (mpe), "protocol", mpe->protocol);
+	g_mime_object_set_content_type_parameter (GMIME_OBJECT (mpe), "protocol",
+						  ctx->encrypt_protocol);
 	g_mime_multipart_set_boundary (GMIME_MULTIPART (mpe), NULL);
 	
 	return 0;
@@ -340,9 +334,9 @@ g_mime_multipart_encrypted_decrypt (GMimeMultipartEncrypted *mpe, GMimeCipherCon
 				    GError **err)
 {
 	GMimeObject *decrypted, *version, *encrypted;
-	const GMimeContentType *mime_type;
 	GMimeStream *stream, *ciphertext;
 	GMimeStream *filtered_stream;
+	GMimeContentType *mime_type;
 	GMimeSignatureValidity *sv;
 	GMimeDataWrapper *wrapper;
 	GMimeFilter *crlf_filter;
@@ -356,7 +350,6 @@ g_mime_multipart_encrypted_decrypt (GMimeMultipartEncrypted *mpe, GMimeCipherCon
 	
 	if (mpe->decrypted) {
 		/* we seem to have already decrypted the part */
-		g_object_ref (mpe->decrypted);
 		return mpe->decrypted;
 	}
 	
@@ -383,7 +376,6 @@ g_mime_multipart_encrypted_decrypt (GMimeMultipartEncrypted *mpe, GMimeCipherCon
 		g_set_error (err, GMIME_ERROR, GMIME_ERROR_PROTOCOL_ERROR,
 			     "Failed to decrypt MIME part: protocol error");
 		
-		g_object_unref (version);
 		g_free (content_type);
 		
 		return NULL;
@@ -395,7 +387,6 @@ g_mime_multipart_encrypted_decrypt (GMimeMultipartEncrypted *mpe, GMimeCipherCon
 	mime_type = g_mime_object_get_content_type (encrypted);
 	if (!g_mime_content_type_is_type (mime_type, "application", "octet-stream")) {
 		g_object_unref (encrypted);
-		g_object_unref (version);
 		return NULL;
 	}
 	
@@ -403,7 +394,6 @@ g_mime_multipart_encrypted_decrypt (GMimeMultipartEncrypted *mpe, GMimeCipherCon
 	wrapper = g_mime_part_get_content_object (GMIME_PART (encrypted));
 	ciphertext = g_mime_data_wrapper_get_stream (wrapper);
 	g_mime_stream_reset (ciphertext);
-	g_object_unref (wrapper);
 	
 	stream = g_mime_stream_mem_new ();
 	filtered_stream = g_mime_stream_filter_new (stream);
@@ -414,7 +404,6 @@ g_mime_multipart_encrypted_decrypt (GMimeMultipartEncrypted *mpe, GMimeCipherCon
 	/* get the cleartext */
 	if (!(sv = g_mime_cipher_context_decrypt (ctx, ciphertext, filtered_stream, err))) {
 		g_object_unref (filtered_stream);
-		g_object_unref (ciphertext);
 		g_object_unref (stream);
 		
 		return NULL;
@@ -422,7 +411,6 @@ g_mime_multipart_encrypted_decrypt (GMimeMultipartEncrypted *mpe, GMimeCipherCon
 	
 	g_mime_stream_flush (filtered_stream);
 	g_object_unref (filtered_stream);
-	g_object_unref (ciphertext);
 	
 	g_mime_stream_reset (stream);
 	parser = g_mime_parser_new ();
@@ -442,7 +430,6 @@ g_mime_multipart_encrypted_decrypt (GMimeMultipartEncrypted *mpe, GMimeCipherCon
 	}
 	
 	/* cache the decrypted part */
-	g_object_ref (decrypted);
 	mpe->decrypted = decrypted;
 	mpe->validity = sv;
 	
