@@ -140,7 +140,6 @@ static void
 g_mime_multipart_init (GMimeMultipart *multipart, GMimeMultipartClass *klass)
 {
 	multipart->children = g_ptr_array_new ();
-	multipart->boundary = NULL;
 	multipart->preface = NULL;
 	multipart->postface = NULL;
 }
@@ -151,7 +150,6 @@ g_mime_multipart_finalize (GObject *object)
 	GMimeMultipart *multipart = (GMimeMultipart *) object;
 	guint i;
 	
-	g_free (multipart->boundary);
 	g_free (multipart->preface);
 	g_free (multipart->postface);
 	
@@ -219,13 +217,6 @@ multipart_remove_header (GMimeObject *object, const char *header)
 static void
 multipart_set_content_type (GMimeObject *object, GMimeContentType *content_type)
 {
-	GMimeMultipart *multipart = (GMimeMultipart *) object;
-	const char *boundary;
-	
-	boundary = g_mime_content_type_get_parameter (content_type, "boundary");
-	g_free (multipart->boundary);
-	multipart->boundary = g_strdup (boundary);
-	
 	GMIME_OBJECT_CLASS (parent_class)->set_content_type (object, content_type);
 }
 
@@ -240,6 +231,7 @@ multipart_write_to_stream (GMimeObject *object, GMimeStream *stream)
 {
 	GMimeMultipart *multipart = (GMimeMultipart *) object;
 	ssize_t nwritten, total = 0;
+	const char *boundary;
 	GMimeObject *part;
 	guint i;
 	
@@ -247,8 +239,11 @@ multipart_write_to_stream (GMimeObject *object, GMimeStream *stream)
 	 * header (in which case it should already be set... or if
 	 * not, then it's a broken multipart and so we don't want to
 	 * alter it or we'll completely break the output) */
-	if (!multipart->boundary && !g_mime_header_list_has_raw (object->headers))
+	boundary = g_mime_object_get_content_type_parameter (object, "boundary");
+	if (!boundary && !g_mime_header_list_has_raw (object->headers)) {
 		g_mime_multipart_set_boundary (multipart, NULL);
+		boundary = g_mime_object_get_content_type_parameter (object, "boundary");
+	}
 	
 	/* write the content headers */
 	if ((nwritten = g_mime_header_list_write_to_stream (object->headers, stream)) == -1)
@@ -274,7 +269,7 @@ multipart_write_to_stream (GMimeObject *object, GMimeStream *stream)
 		part = multipart->children->pdata[i];
 		
 		/* write the boundary */
-		if ((nwritten = g_mime_stream_printf (stream, "\n--%s\n", multipart->boundary)) == -1)
+		if ((nwritten = g_mime_stream_printf (stream, "\n--%s\n", boundary)) == -1)
 			return -1;
 		
 		total += nwritten;
@@ -287,8 +282,8 @@ multipart_write_to_stream (GMimeObject *object, GMimeStream *stream)
 	}
 	
 	/* write the end-boundary (but only if a boundary is set) */
-	if (multipart->boundary) {
-		if ((nwritten = g_mime_stream_printf (stream, "\n--%s--\n", multipart->boundary)) == -1)
+	if (boundary) {
+		if ((nwritten = g_mime_stream_printf (stream, "\n--%s--\n", boundary)) == -1)
 			return -1;
 		
 		total += nwritten;
@@ -763,9 +758,6 @@ multipart_set_boundary (GMimeMultipart *multipart, const char *boundary)
 		boundary = bbuf;
 	}
 	
-	g_free (multipart->boundary);
-	multipart->boundary = g_strdup (boundary);
-	
 	g_mime_object_set_content_type_parameter (GMIME_OBJECT (multipart), "boundary", boundary);
 }
 
@@ -790,10 +782,15 @@ g_mime_multipart_set_boundary (GMimeMultipart *multipart, const char *boundary)
 static const char *
 multipart_get_boundary (GMimeMultipart *multipart)
 {
-	if (!multipart->boundary)
-		multipart_set_boundary (multipart, NULL);
+	GMimeObject *object = (GMimeObject *) multipart;
+	const char *boundary;
 	
-	return multipart->boundary;
+	if ((boundary = g_mime_object_get_content_type_parameter (object, "boundary")))
+		return boundary;
+	
+	multipart_set_boundary (multipart, NULL);
+	
+	return g_mime_object_get_content_type_parameter (object, "boundary");
 }
 
 
@@ -816,27 +813,37 @@ g_mime_multipart_get_boundary (GMimeMultipart *multipart)
 }
 
 
-/**
- * g_mime_multipart_foreach: 
- * @multipart: a multipart
- * @callback: function to call for @multipart and all of its subparts
- * @user_data: extra data to pass to the callback
- * 
- * Calls @callback on each of @multipart's subparts.
- **/
-void
-g_mime_multipart_foreach (GMimeMultipart *multipart, GMimePartFunc callback, gpointer user_data)
+static void
+multipart_foreach (GMimeMultipart *multipart, GMimeObjectForeachFunc callback, gpointer user_data)
 {
 	GMimeObject *part;
 	guint i;
 	
+	for (i = 0; i < multipart->children->len; i++) {
+		part = (GMimeObject *) multipart->children->pdata[i];
+		callback ((GMimeObject *) multipart, part, user_data);
+		
+		if (GMIME_IS_MULTIPART (part))
+			multipart_foreach ((GMimeMultipart *) part, callback, user_data);
+	}
+}
+
+
+/**
+ * g_mime_multipart_foreach: 
+ * @multipart: a #GMimeMultipart
+ * @callback: function to call for each of @multipart's subparts.
+ * @user_data: user-supplied callback data
+ * 
+ * Recursively calls @callback on each of @multipart's subparts.
+ **/
+void
+g_mime_multipart_foreach (GMimeMultipart *multipart, GMimeObjectForeachFunc callback, gpointer user_data)
+{
 	g_return_if_fail (GMIME_IS_MULTIPART (multipart));
 	g_return_if_fail (callback != NULL);
 	
-	for (i = 0; i < multipart->children->len; i++) {
-		part = multipart->children->pdata[i];
-		callback (part, user_data);
-	}
+	multipart_foreach (multipart, callback, user_data);
 }
 
 
@@ -848,7 +855,7 @@ g_mime_multipart_foreach (GMimeMultipart *multipart, GMimePartFunc callback, gpo
  * Gets the mime part with the content-id @content_id from the
  * multipart @multipart.
  *
- * Returns: the GMimeObject whose content-id matches the search string,
+ * Returns: the #GMimeObject whose content-id matches the search string,
  * or %NULL if a match cannot be found.
  **/
 GMimeObject *
@@ -868,12 +875,13 @@ g_mime_multipart_get_subpart_from_content_id (GMimeMultipart *multipart, const c
 	for (i = 0; i < multipart->children->len; i++) {
 		subpart = multipart->children->pdata[i];
 		
+		if (subpart->content_id && !strcmp (subpart->content_id, content_id))
+			return subpart;
+		
 		if (GMIME_IS_MULTIPART (subpart)) {
 			mpart = (GMimeMultipart *) subpart;
 			if ((part = g_mime_multipart_get_subpart_from_content_id (mpart, content_id)))
 				return part;
-		} else if (subpart->content_id && !strcmp (subpart->content_id, content_id)) {
-			return subpart;
 		}
 	}
 	
