@@ -57,6 +57,13 @@
  **/
 
 
+typedef struct _ContentType {
+	char *type, *subtype;
+	gboolean exists;
+} ContentType;
+
+extern void _g_mime_object_set_content_type (GMimeObject *object, GMimeContentType *content_type);
+
 static void g_mime_parser_class_init (GMimeParserClass *klass);
 static void g_mime_parser_init (GMimeParser *parser, GMimeParserClass *klass);
 static void g_mime_parser_finalize (GObject *object);
@@ -64,9 +71,9 @@ static void g_mime_parser_finalize (GObject *object);
 static void parser_init (GMimeParser *parser, GMimeStream *stream);
 static void parser_close (GMimeParser *parser);
 
-static GMimeObject *parser_construct_leaf_part (GMimeParser *parser, GMimeContentType *content_type,
+static GMimeObject *parser_construct_leaf_part (GMimeParser *parser, ContentType *content_type,
 						int *found);
-static GMimeObject *parser_construct_multipart (GMimeParser *parser, GMimeContentType *content_type,
+static GMimeObject *parser_construct_multipart (GMimeParser *parser, ContentType *content_type,
 						int *found);
 
 static GObjectClass *parent_class = NULL;
@@ -112,8 +119,12 @@ struct _GMimeParserPrivate {
 	char *rawptr;
 	size_t rawleft;
 	
-	gint64 headers_start;
-	gint64 header_start;
+	/* current message headerblock offsets */
+	gint64 headers_begin;
+	gint64 headers_end;
+	
+	/* current header field offset */
+	gint64 header_offset;
 	
 	short int state;
 	
@@ -125,7 +136,6 @@ struct _GMimeParserPrivate {
 	unsigned short int persist_stream:1;
 	unsigned short int respect_content_length:1;
 	
-	GMimeContentType *content_type;
 	struct _header_raw *headers;
 	
 	struct _boundary_stack *bounds;
@@ -325,8 +335,10 @@ parser_init (GMimeParser *parser, GMimeStream *stream)
 	priv->rawptr = priv->rawbuf;
 	priv->rawleft = SCAN_HEAD - 1;
 	
-	priv->headers_start = -1;
-	priv->header_start = -1;
+	priv->headers_begin = -1;
+	priv->headers_end = -1;
+	
+	priv->header_offset = -1;
 	
 	priv->midline = FALSE;
 	priv->seekable = offset != -1;
@@ -396,7 +408,7 @@ g_mime_parser_new_with_stream (GMimeStream *stream)
 
 /**
  * g_mime_parser_init_with_stream:
- * @parser: MIME parser object
+ * @parser: a #GMimeParser context
  * @stream: raw message or part stream
  *
  * Initializes @parser to use @stream.
@@ -428,7 +440,7 @@ g_mime_parser_init_with_stream (GMimeParser *parser, GMimeStream *stream)
 
 /**
  * g_mime_parser_get_persist_stream:
- * @parser: MIME parser object
+ * @parser: a #GMimeParser context
  *
  * Gets whether or not the underlying stream is persistent.
  *
@@ -446,7 +458,7 @@ g_mime_parser_get_persist_stream (GMimeParser *parser)
 
 /**
  * g_mime_parser_set_persist_stream:
- * @parser: MIME parser object
+ * @parser: a #GMimeParser context
  * @persist: persist attribute
  *
  * Sets whether or not the @parser's underlying stream is persistent.
@@ -473,7 +485,7 @@ g_mime_parser_set_persist_stream (GMimeParser *parser, gboolean persist)
 
 /**
  * g_mime_parser_get_scan_from:
- * @parser: MIME parser object
+ * @parser: a #GMimeParser context
  *
  * Gets whether or not @parser is set to scan mbox-style From-lines.
  *
@@ -491,7 +503,7 @@ g_mime_parser_get_scan_from (GMimeParser *parser)
 
 /**
  * g_mime_parser_set_scan_from:
- * @parser: MIME parser object
+ * @parser: a #GMimeParser context
  * @scan_from: %TRUE to scan From-lines or %FALSE otherwise
  *
  * Sets whether or not @parser should scan mbox-style From-lines.
@@ -507,7 +519,7 @@ g_mime_parser_set_scan_from (GMimeParser *parser, gboolean scan_from)
 
 /**
  * g_mime_parser_get_respect_content_length:
- * @parser: MIME parser object
+ * @parser: a #GMimeParser context
  *
  * Gets whether or not @parser is set to use Content-Length for
  * determining the offset of the end of the message.
@@ -526,7 +538,7 @@ g_mime_parser_get_respect_content_length (GMimeParser *parser)
 
 /**
  * g_mime_parser_set_respect_content_length:
- * @parser: MIME parser object
+ * @parser: a #GMimeParser context
  * @respect_content_length: %TRUE if the parser should use Content-Length headers or %FALSE otherwise.
  *
  * Sets whether or not @parser should respect Content-Length headers
@@ -547,7 +559,7 @@ g_mime_parser_set_respect_content_length (GMimeParser *parser, gboolean respect_
 
 /**
  * g_mime_parser_set_header_regex:
- * @parser: MIME parser object
+ * @parser: a #GMimeParser context
  * @regex: regular expression
  * @header_cb: callback function
  * @user_data: user data
@@ -642,7 +654,7 @@ parser_offset (struct _GMimeParserPrivate *priv, const char *inptr)
 
 /**
  * g_mime_parser_tell:
- * @parser: MIME parser object
+ * @parser: a #GMimeParser context
  *
  * Gets the current stream offset from the parser's internal stream.
  *
@@ -812,7 +824,7 @@ header_parse (GMimeParser *parser, struct _header_raw **tail)
 	if (*inptr != ':') {
 		/* ignore invalid headers */
 		w(g_warning ("Invalid header at %lld: '%s'",
-			     (long long) priv->header_start,
+			     (long long) priv->header_offset,
 			     priv->headerbuf));
 		
 		priv->headerleft += priv->headerptr - priv->headerbuf;
@@ -844,7 +856,7 @@ header_parse (GMimeParser *parser, struct _header_raw **tail)
 	
 	header->value = g_strndup (start, end - start);
 	
-	header->offset = priv->header_start;
+	header->offset = priv->header_offset;
 	
 	(*tail)->next = header;
 	*tail = header;
@@ -874,8 +886,10 @@ parser_step_headers (GMimeParser *parser)
 	raw_header_reset (priv);
 	header_raw_clear (&priv->headers);
 	hend = (struct _header_raw *) &priv->headers;
-	priv->headers_start = parser_offset (priv, NULL);
-	priv->header_start = parser_offset (priv, NULL);
+	priv->header_offset = parser_offset (priv, NULL);
+	
+	if (priv->headers_begin == -1)
+		priv->headers_begin = parser_offset (priv, NULL);
 	
 	inptr = priv->inptr;
 	inend = priv->inend;
@@ -898,7 +912,7 @@ parser_step_headers (GMimeParser *parser)
 			/* if we are scanning a new line, check for a folded header */
 			if (!priv->midline && continuation && (*inptr != ' ' && *inptr != '\t')) {
 				header_parse (parser, &hend);
-				priv->header_start = parser_offset (priv, inptr);
+				priv->header_offset = parser_offset (priv, inptr);
 				continuation = FALSE;
 				fieldname = TRUE;
 				valid = TRUE;
@@ -1006,6 +1020,9 @@ parser_step_headers (GMimeParser *parser)
 	if (priv->headerptr > priv->headerbuf)
 		header_parse (parser, &hend);
 	
+	if (priv->headers_end == -1)
+		priv->headers_end = parser_offset (priv, start);
+	
 	priv->state = GMIME_PARSER_STATE_HEADERS_END;
 	*priv->rawptr = '\0';
 	priv->inptr = inptr;
@@ -1014,6 +1031,9 @@ parser_step_headers (GMimeParser *parser)
 	
  next_message:
 	
+	if (priv->headers_end == -1)
+		priv->headers_end = parser_offset (priv, start);
+	
 	priv->state = GMIME_PARSER_STATE_COMPLETE;
 	*priv->rawptr = '\0';
 	priv->inptr = start;
@@ -1022,6 +1042,9 @@ parser_step_headers (GMimeParser *parser)
 	
  content_start:
 	
+	if (priv->headers_end == -1)
+		priv->headers_end = parser_offset (priv, start);
+	
 	priv->state = GMIME_PARSER_STATE_CONTENT;
 	*priv->rawptr = '\0';
 	priv->inptr = start;
@@ -1029,16 +1052,49 @@ parser_step_headers (GMimeParser *parser)
 	return 0;
 }
 
-static GMimeContentType *
+static void
+content_type_destroy (ContentType *content_type)
+{
+	g_free (content_type->subtype);
+	g_free (content_type->type);
+	g_free (content_type);
+}
+
+static gboolean
+content_type_is_type (ContentType *content_type, const char *type, const char *subtype)
+{
+	if (!strcmp (type, "*") || !g_ascii_strcasecmp (content_type->type, type)) {
+		if (!strcmp (subtype, "*")) {
+			/* special case */
+			return TRUE;
+		}
+		
+		if (!g_ascii_strcasecmp (content_type->subtype, subtype))
+			return TRUE;
+	}
+	
+	return FALSE;
+}
+
+static ContentType *
 parser_content_type (GMimeParser *parser)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
-	const char *content_type;
+	register const char *inptr;
+	ContentType *content_type;
+	const char *value;
 	
-	if ((content_type = header_raw_find (priv->headers, "Content-Type", NULL)))
-		return g_mime_content_type_new_from_string (content_type);
+	content_type = g_new (ContentType, 1);
 	
-	return NULL;
+	if (!(value = header_raw_find (priv->headers, "Content-Type", NULL)) ||
+	    !g_mime_parse_content_type (&value, &content_type->type, &content_type->subtype)) {
+		content_type->type = g_strdup ("text");
+		content_type->subtype = g_strdup ("plain");
+	}
+	
+	content_type->exists = value != NULL;
+	
+	return content_type;
 }
 
 static int
@@ -1085,12 +1141,16 @@ parser_step (GMimeParser *parser)
 	case GMIME_PARSER_STATE_ERROR:
 		break;
 	case GMIME_PARSER_STATE_INIT:
+		parser->priv->headers_begin = -1;
+		parser->priv->headers_end = -1;
 		if (priv->scan_from)
 			priv->state = GMIME_PARSER_STATE_FROM;
 		else
 			priv->state = GMIME_PARSER_STATE_HEADERS;
 		break;
 	case GMIME_PARSER_STATE_FROM:
+		parser->priv->headers_begin = -1;
+		parser->priv->headers_end = -1;
 		parser_step_from (parser);
 		break;
 	case GMIME_PARSER_STATE_HEADERS:
@@ -1320,8 +1380,8 @@ static void
 parser_scan_message_part (GMimeParser *parser, GMimeMessagePart *mpart, int *found)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
-	GMimeContentType *content_type;
 	struct _header_raw *header;
+	ContentType *content_type;
 	GMimeMessage *message;
 	GMimeObject *object;
 	
@@ -1343,14 +1403,13 @@ parser_scan_message_part (GMimeParser *parser, GMimeMessagePart *mpart, int *fou
 		header = header->next;
 	}
 	
-	if (!(content_type = parser_content_type (parser)))
-		content_type = g_mime_content_type_new ("text", "plain");
-	
-	if (g_mime_content_type_is_type (content_type, "multipart", "*"))
+	content_type = parser_content_type (parser);
+	if (content_type_is_type (content_type, "multipart", "*"))
 		object = parser_construct_multipart (parser, content_type, found);
 	else
 		object = parser_construct_leaf_part (parser, content_type, found);
 	
+	content_type_destroy (content_type);
 	message->mime_part = object;
 	
 	g_mime_message_part_set_message (mpart, message);
@@ -1358,7 +1417,7 @@ parser_scan_message_part (GMimeParser *parser, GMimeMessagePart *mpart, int *fou
 }
 
 static GMimeObject *
-parser_construct_leaf_part (GMimeParser *parser, GMimeContentType *content_type, int *found)
+parser_construct_leaf_part (GMimeParser *parser, ContentType *content_type, int *found)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
 	struct _header_raw *header;
@@ -1368,6 +1427,14 @@ parser_construct_leaf_part (GMimeParser *parser, GMimeContentType *content_type,
 	
 	object = g_mime_object_new_type (content_type->type, content_type->subtype);
 	
+	if (!content_type->exists) {
+		GMimeContentType *mime_type;
+		
+		mime_type = g_mime_content_type_new ("text", "plain");
+		_g_mime_object_set_content_type (object, mime_type);
+		g_object_unref (mime_type);
+	}
+	
 	header = priv->headers;
 	while (header) {
 		g_mime_object_append_header (object, header->name, header->value);
@@ -1376,9 +1443,6 @@ parser_construct_leaf_part (GMimeParser *parser, GMimeContentType *content_type,
 	
 	header_raw_clear (&priv->headers);
 	
-	g_mime_object_set_content_type (object, content_type);
-	g_object_unref (content_type);
-	
 	g_mime_header_list_set_raw (object->headers, priv->rawbuf);
 	raw_header_reset (priv);
 	
@@ -1468,7 +1532,7 @@ static int
 parser_scan_multipart_subparts (GMimeParser *parser, GMimeMultipart *multipart)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
-	GMimeContentType *content_type;
+	ContentType *content_type;
 	GMimeObject *subpart;
 	int found;
 	
@@ -1491,15 +1555,14 @@ parser_scan_multipart_subparts (GMimeParser *parser, GMimeMultipart *multipart)
 			break;
 		}
 		
-		if (!(content_type = parser_content_type (parser)))
-			content_type = g_mime_content_type_new ("text", "plain");
-		
-		if (g_mime_content_type_is_type (content_type, "multipart", "*"))
+		content_type = parser_content_type (parser);
+		if (content_type_is_type (content_type, "multipart", "*"))
 			subpart = parser_construct_multipart (parser, content_type, &found);
 		else
 			subpart = parser_construct_leaf_part (parser, content_type, &found);
 		
 		g_mime_multipart_add (multipart, subpart);
+		content_type_destroy (content_type);
 		g_object_unref (subpart);
 	} while (found == FOUND_BOUNDARY && found_immediate_boundary (priv, FALSE));
 	
@@ -1507,7 +1570,7 @@ parser_scan_multipart_subparts (GMimeParser *parser, GMimeMultipart *multipart)
 }
 
 static GMimeObject *
-parser_construct_multipart (GMimeParser *parser, GMimeContentType *content_type, int *found)
+parser_construct_multipart (GMimeParser *parser, ContentType *content_type, int *found)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
 	struct _header_raw *header;
@@ -1527,9 +1590,6 @@ parser_construct_multipart (GMimeParser *parser, GMimeContentType *content_type,
 	
 	header_raw_clear (&priv->headers);
 	
-	g_mime_object_set_content_type (object, content_type);
-	g_object_unref (content_type);
-	
 	g_mime_header_list_set_raw (object->headers, priv->rawbuf);
 	raw_header_reset (priv);
 	
@@ -1543,7 +1603,7 @@ parser_construct_multipart (GMimeParser *parser, GMimeContentType *content_type,
 		}
 	}
 	
-	boundary = g_mime_content_type_get_parameter (content_type, "boundary");
+	boundary = g_mime_object_get_content_type_parameter (object, "boundary");
 	if (boundary) {
 		parser_push_boundary (parser, boundary);
 		
@@ -1573,7 +1633,7 @@ static GMimeObject *
 parser_construct_part (GMimeParser *parser)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
-	GMimeContentType *content_type;
+	ContentType *content_type;
 	GMimeObject *object;
 	int found;
 	
@@ -1583,21 +1643,21 @@ parser_construct_part (GMimeParser *parser)
 			return NULL;
 	}
 	
-	if (!(content_type = parser_content_type (parser)))
-		content_type = g_mime_content_type_new ("text", "plain");
-	
-	if (g_mime_content_type_is_type (content_type, "multipart", "*"))
+	content_type = parser_content_type (parser);
+	if (content_type_is_type (content_type, "multipart", "*"))
 		object = parser_construct_multipart (parser, content_type, &found);
 	else
 		object = parser_construct_leaf_part (parser, content_type, &found);
 	
+	content_type_destroy (content_type);
+	
 	return object;
 }
 
 
 /**
  * g_mime_parser_construct_part:
- * @parser: MIME parser object
+ * @parser: a #GMimeParser context
  *
  * Constructs a MIME part from @parser.
  *
@@ -1617,8 +1677,8 @@ parser_construct_message (GMimeParser *parser)
 {
 	struct _GMimeParserPrivate *priv = parser->priv;
 	unsigned long content_length = ULONG_MAX;
-	GMimeContentType *content_type;
 	struct _header_raw *header;
+	ContentType *content_type;
 	GMimeMessage *message;
 	GMimeObject *object;
 	char *endptr;
@@ -1655,14 +1715,13 @@ parser_construct_message (GMimeParser *parser)
 			priv->bounds->content_end = parser_offset (priv, NULL) + content_length;
 	}
 	
-	if (!(content_type = parser_content_type (parser)))
-		content_type = g_mime_content_type_new ("text", "plain");
-	
-	if (content_type && g_mime_content_type_is_type (content_type, "multipart", "*"))
+	content_type = parser_content_type (parser);
+	if (content_type_is_type (content_type, "multipart", "*"))
 		object = parser_construct_multipart (parser, content_type, &found);
 	else
 		object = parser_construct_leaf_part (parser, content_type, &found);
 	
+	content_type_destroy (content_type);
 	message->mime_part = object;
 	
 	if (priv->scan_from) {
@@ -1676,7 +1735,7 @@ parser_construct_message (GMimeParser *parser)
 
 /**
  * g_mime_parser_construct_message:
- * @parser: MIME parser object
+ * @parser: a #GMimeParser context
  *
  * Constructs a MIME message from @parser.
  *
@@ -1693,7 +1752,7 @@ g_mime_parser_construct_message (GMimeParser *parser)
 
 /**
  * g_mime_parser_get_from:
- * @parser: MIME parser object
+ * @parser: a #GMimeParser context
  *
  * Gets the mbox-style From-line of the most recently parsed message
  * (gotten from g_mime_parser_construct_message()).
@@ -1721,13 +1780,13 @@ g_mime_parser_get_from (GMimeParser *parser)
 
 /**
  * g_mime_parser_get_from_offset:
- * @parser: MIME parser object
+ * @parser: a #GMimeParser context
  *
  * Gets the offset of the most recently parsed mbox-style From-line
  * (gotten from g_mime_parser_construct_message()).
  *
  * Returns: the offset of the most recently parsed mbox-style From-line
- * or -1 on error.
+ * or %-1 on error.
  **/
 gint64
 g_mime_parser_get_from_offset (GMimeParser *parser)
@@ -1742,3 +1801,41 @@ g_mime_parser_get_from_offset (GMimeParser *parser)
 	
 	return priv->from_offset;
 }
+
+
+/**
+ * g_mime_parser_get_headers_begin:
+ * @parser: a #GMimeParser context
+ *
+ * Gets the stream offset of the beginning of the headers of the most
+ * recently parsed message.
+ *
+ * Returns: the offset of the beginning of the headers of the most
+ * recently parsed message or %-1 on error.
+ **/
+gint64
+g_mime_parser_get_headers_begin (GMimeParser *parser)
+{
+	g_return_val_if_fail (GMIME_IS_PARSER (parser), -1);
+	
+	return parser->priv->headers_begin;
+}
+
+
+/**
+ * g_mime_parser_get_headers_end:
+ * @parser: a #GMimeParser context
+ *
+ * Gets the stream offset of the end of the headers of the most
+ * recently parsed message.
+ *
+ * Returns: the offset of the end of the headers of the most recently
+ * parsed message or %-1 on error.
+ **/
+gint64
+g_mime_parser_get_headers_end (GMimeParser *parser)
+{
+	g_return_val_if_fail (GMIME_IS_PARSER (parser), -1);
+	
+	return parser->priv->headers_end;
+}
