@@ -37,6 +37,14 @@ G_BEGIN_DECLS
 typedef struct _GMimeStreamFilter GMimeStreamFilter;
 typedef struct _GMimeStreamFilterClass GMimeStreamFilterClass;
 
+/**
+ * GMimeStreamFilter:
+ * @parent_object: parent #GMimeStream
+ * @priv: private state data
+ * @source: source stream
+ *
+ * A #GMimeStream which passes data through any #GMimeFilter objects.
+ **/
 struct _GMimeStreamFilter {
 	GMimeStream parent_object;
 	
@@ -53,7 +61,7 @@ struct _GMimeStreamFilterClass {
 
 GType g_mime_stream_filter_get_type (void);
 
-GMimeStream *g_mime_stream_filter_new_with_stream (GMimeStream *stream);
+GMimeStream *g_mime_stream_filter_new (GMimeStream *stream);
 
 int g_mime_stream_filter_add (GMimeStreamFilter *fstream, GMimeFilter *filter);
 void g_mime_stream_filter_remove (GMimeStreamFilter *fstream, int id);
