@@ -60,6 +60,10 @@ static void
 print_mime_struct (GMimeStream *stream, GMimeObject *part, int depth)
 {
 	const GMimeContentType *type;
+	GMimeMultipart *multipart;
+	GMimeMessagePart *mpart;
+	GMimeObject *subpart;
+	int i, n;
 	
 	print_depth (stream, depth);
 	
@@ -68,15 +72,15 @@ print_mime_struct (GMimeStream *stream, GMimeObject *part, int depth)
 	g_mime_stream_printf (stream, "Content-Type: %s/%s\n", type->type, type->subtype);
 	
 	if (GMIME_IS_MULTIPART (part)) {
-		GList *subpart;
+		multipart = (GMimeMultipart *) part;
 		
-		subpart = GMIME_MULTIPART (part)->subparts;
-		while (subpart) {
-			print_mime_struct (stream, subpart->data, depth + 1);
-			subpart = subpart->next;
+		n = g_mime_multipart_get_count (multipart);
+		for (i = 0; i < n; i++) {
+			subpart = g_mime_multipart_get_part (multipart, i);
+			print_mime_struct (stream, subpart, depth + 1);
 		}
 	} else if (GMIME_IS_MESSAGE_PART (part)) {
-		GMimeMessagePart *mpart = (GMimeMessagePart *) part;
+		mpart = (GMimeMessagePart *) part;
 		
 		if (mpart->message)
 			print_mime_struct (stream, mpart->message->mime_part, depth + 1);
@@ -95,9 +99,9 @@ header_cb (GMimeParser *parser, const char *header, const char *value, gint64 of
 static void
 test_parser (GMimeParser *parser, GMimeStream *mbox, GMimeStream *summary)
 {
+	gint64 headers_begin, headers_end, start, end;
 	GMimeMessage *message;
 	const char *exev;
-	gint64 start, end;
 	int nmsg = 0;
 	char *from;
 	
@@ -108,11 +112,16 @@ test_parser (GMimeParser *parser, GMimeStream *mbox, GMimeStream *summary)
 		
 		end = g_mime_parser_tell (parser);
 		
+		headers_begin = g_mime_parser_get_headers_begin (parser);
+		headers_end = g_mime_parser_get_headers_end (parser);
+		
 		g_mime_stream_printf (summary, "message offsets: %lld, %lld\n", start, end);
+		g_mime_stream_printf (summary, "header offsets: %lld, %lld\n",
+				      headers_begin, headers_end);
 		
 		from = g_mime_parser_get_from (parser);
 		g_mime_stream_printf (summary, "%s\n", from);
-		exev = g_mime_message_get_header (message, "X-Evolution");
+		exev = g_mime_object_get_header ((GMimeObject *) message, "X-Evolution");
 		g_mime_stream_printf (summary, "X-Evolution: %s\n", exev ? exev : "None");
 		print_mime_struct (summary, message->mime_part, 0);
 		g_mime_stream_write (summary, "\n", 1);
@@ -415,6 +424,7 @@ int main (int argc, char **argv)
 		
 		g_object_unref (istream);
 		g_object_unref (ostream);
+		g_object_unref (parser);
 		
 #ifdef ENABLE_MBOX_MATCH
 		g_object_unref (mstream);
