@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
- *  Copyright (C) 2000-2009 Jeffrey Stedfast
+ *  Copyright (C) 2000-2008 Jeffrey Stedfast
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -150,6 +150,22 @@ static struct {
 	{ "undisclosed-recipients",
 	  "undisclosed-recipients",
 	  "undisclosed-recipients" },
+	/* The following 2 addr-specs are invalid according to the
+	 * spec, but apparently some japanese cellphones use them?
+	 * See Evolution bug #547969 */
+	{ "some...dots@hocus.pocus.net",
+	  "some...dots@hocus.pocus.net",
+	  "some...dots@hocus.pocus.net" },
+	{ "some.dots..@hocus.pocus.net",
+	  "some.dots..@hocus.pocus.net",
+	  "some.dots..@hocus.pocus.net" },
+	/* The following test case is to check that we properly handle
+	 * mailbox addresses that do not have any lwsp between the
+	 * name component and the addr-spec. See Evolution bug
+	 * #347520 */
+	{ "Canonical Patch Queue Manager<pqm@pqm.ubuntu.com>",
+	  "Canonical Patch Queue Manager <pqm@pqm.ubuntu.com>",
+	  "Canonical Patch Queue Manager <pqm@pqm.ubuntu.com>" },
 };
 
 static void
@@ -165,7 +181,7 @@ test_addrspec (void)
 		
 		testsuite_check ("addrspec[%u]", i);
 		try {
-			if (!(addrlist = internet_address_parse_string (addrspec[i].input)))
+			if (!(addrlist = internet_address_list_parse_string (addrspec[i].input)))
 				throw (exception_new ("could not parse addr-spec"));
 			
 			str = internet_address_list_to_string (addrlist, FALSE);
@@ -184,7 +200,7 @@ test_addrspec (void)
 		
 		g_free (str);
 		if (addrlist)
-			internet_address_list_destroy (addrlist);
+			g_object_unref (addrlist);
 	}
 }
 
@@ -213,6 +229,9 @@ static struct {
 	{ "Sat, 24 Mar 2007 21:23:03 GMT",
 	  "Sat, 24 Mar 2007 21:23:03 +0000",
 	  1174771383, 0 },
+	{ "17-6-2008 17:10:08",
+	  "Tue, 17 Jun 2008 17:10:08 +0000",
+	  1213722608, 0 },
 };
 
 static void
